<?php
    // A better var dump. This will terminate the script following execution.
    function nvar_dump($data) {
        var_dump($data); exit;
    }

    // A better redirect.
    function redirect($path, $dir = true) {
        global $workurl;
        global $displaypath;

        if($dir)
            $path = $workurl . $path;

        // Do the redirect, then prevent further page execution.
        header('Location: ' . $path);
        exit;
    }

    // Sets a number of default parameters.
    function initialise() {
        setDisplayPath();
        setFormatter();
        loadPolicies();
        setDefinitions();
    }

    // Loads definitions.
    function setDefinitions() {
        global $workpath, $defs;
        require($workpath . '/assets/data/defs.php');
    }

    // Self-explanatory. Loads them into memory. Does nothing else.
    function loadPolicies() {
        global $workpath;
        require($workpath . '/assets/data/disclaimer.php');
        require($workpath . '/assets/data/privacy.php');
    }

    // Sets the filepath and SEO data paths.
    function setDisplayPath() {
        global $workpath, $workurl, $sitedir, $pages, $navitems;
        global $parameters, $displaypath, $datapath, $pagedata;

        // Default parameters.
        $displaypath = '/home';
        $datapath = '/';

        // If there are URL parameters, we need to process those.
        if($parampos = strpos($_SERVER['REQUEST_URI'], '?')) {
            $datapath .= substr($_SERVER['REQUEST_URI'], $parampos);

            // We want to make sure external pages on the redirector
            //     are properly verified and ::: is unlikely to occur
            //     in paths of other webpages, so we use this as a
            //     delimeter for the redirector.
            if(strpos($datapath, ':::')) {
                // Split the parameters out, based on placement of the first ::: in the string.
                //     Anything before the first ::: will be a page address or path, not a
                //     parameter, thus we won't even need it. Anything after and including will
                //     be a parameter of some sort.
                $pos = strpos($datapath, ':::');
                $parameters = array_filter(explode(':::', substr($datapath, $pos)));

                // For each detected parameter,
                foreach($parameters as $key => $param) {

                    // If there isn't a value with the parameter, discard it and skip this loop.
                    if(strpos($param, "=") == false) {
                        unset($parameters[$key]);
                        continue;
                    }

                    // Three-step process for setting a new parameter.
                    //    1. Get the position of the delimeter of key and value.
                    //    2. Set the new key to what's *before* that delimeter.
                    //    3. Set the new value to what's after.
                    $pos2 = strpos($param, "=");
                    $newkey   = strtolower(substr($param, 0, $pos2));
                    $newparam = substr($param, ($pos2+1));

                    // Unset the parameter.
                    unset($parameters[$key]);

                    // If the new key or parameter is NOT empty, set the new paramter in the array.
                    if(!empty($newkey) && !empty($newparam))
                        $parameters[$newkey] = $newparam;

                    // Clear these variables so the next loop has no accidentally-retained values.
                    unset($pos2);
                    unset($newkey);
                    unset($newparam);
                }

                // Set the datapath to what's before the first ::: from the URL.
                // After, unset $pos as we don't need it anymore.
                $datapath = substr($datapath, 0, $pos);
                unset($pos);
            }

            // nvar_dump($datapath);

            // If there is more than one question mark in the datapath, 
            //     remove it and redirect to the new page.
            if(substr_count($datapath, '?') > 1) {
                $exp = explode("?", $datapath);
                redirect($exp[0] . "?" . $exp[1]);
            }

            // Only allow one parameter, so redirect if attempts to insert extra ones are made.
            if(strpos($datapath, '&'))
                redirect(explode('&', $datapath)[0]);

            // Don't end on a forward slash.
            if(substr($datapath, -1) == "/")
                redirect(substr($datapath, 0, -1));                


            // Set display path base.
            $displaypath = '/' . substr($datapath, (strpos($datapath, "?") + 1));
        }

        // Reconfirm the display path.
        $displaypath = $workpath . $displaypath;

        // If we're a directory, we're a directory. Assume home since we're a raw directory
        //     and not a file.
        if(is_dir($displaypath))
            $displaypath .= '/home';

        // Add .php to it, since we want a PHP file in the end.
        $displaypath .= '.php';

        // If the file doesn't exist, show the 404 page and set paths to that.
        if(!file_exists($displaypath) || !in_array($datapath, $navitems['pages'])) {
            $displaypath = $workpath . '/404.php';
            $datapath = "/?404";
        }

        // If we are the 404 page, set a 404 code. This will also allow manually navigating to
        //     the 404 page to show a 404 error.
        if($datapath == "/?404")
            http_response_code(404);

        // Get the pagedata and set it.
        $pagedata = $pages[$datapath];
    }

    // Sets the formatters for dates and times.
    function setFormatter() {
        global $date_formats;
        global $formatter;

        foreach($date_formats as $key => $format) {
            $formatter[$key] = datefmt_create(
                'en_GB',
                IntlDateFormatter::NONE,
                IntlDateFormatter::NONE,
                new DateTimeZone('UTC'), // set UTC, important.
                IntlDateFormatter::GREGORIAN,
                $format
            );
        }
    }

    // Gets the modified time of a file.
    function getModified($path) {
        global $formatter;

        // Sometimes a time will be given, and in that case just return it.
        $time = is_int($path) ? $path : filemtime($path);

        // Process the timestamp into an array based on the formatters set.
        $times = array();
        foreach($formatter as $key => $format)
            $times[$key] = datefmt_format($format, $time);

        return $times;
    }

    // This function does the bulk of the page display.
    function renderPage() {
        global
            $displaypath, $datapath, $workpath,
            $workurl, $sitedir, $navitems,
            $thispage, $custom_modified,
            $parameters, $site_years,
            $disclaimer, $privacy,
            $formatter, $contact_enabled,
            $renderpage, $pagedata, $downloadFile;

        // First, we log a hit and *then* we load the hit data.
        //     Loading page hit data is done here, rather than later or earlier, because
        //     this way it loads the hit we just logged too and doesn't slow down rendering
        //     later in the process
        logHit();
        $page_hitdata = readHits();

        // Pre-set the download flag to false.
        $downloadFile = false;

        // This buffers the page. All parameters needed for the page have been pre-initialised
        //     or otherwise loaded by this point. We capture the output so we can use it later in
        //     the rendering process.
        //
        // In the case of downloading a site backup, the page included will set $downloadFile
        //     and will trigger a download rather than displaying the page.
        $renderpage = true;
        ob_start();
            include($displaypath);
        $page = ob_get_clean();
        $renderpage = false;

        // If we want to download the site,
        if($downloadFile !== false) {

            // Create the site download file and store it's name.
            $downloadFile = createSiteDownload();

            // If the file exists (in rare cases it might bug out), set headers and send the file,
            //     then delete the file and terminate script execution.
            if(file_exists($downloadFile)) {
                // If we want to download a specific file, set headers            
                header('Content-Description: File Transfer');
                header('Content-Type: application/zip');
                header('Content-Disposition: attachment; filename="' . basename($downloadFile) . '"');
                header('Content-Transfer-Encoding: binary');
                header('Expires: 0');
                header('Cache-Control: must-revalidate, post-check=0, pre-check=0');
                header('Pragma: public');
                header('Content-Length: ' . filesize($downloadFile));

                // Clear some things
                ob_clean();
                flush();

                // Then send the file and quit.
                readfile($downloadFile);

                unlink($downloadFile);

                exit;
            }
        }

        // Count words on the page content and get the time that the page was modified.
        //     Some pages use data from other sources, and may have to set their own
        //     modified times as those data sources — and also page content — might be
        //     more recent than the page's code. We always want to use the custom-set
        //     modified time if it has been set.
        $wordcount = countWords($page);
        $getmod =  isset($custom_modified) ? $custom_modified : $displaypath;
        $modified = getModified($getmod);

        // Load in the full page into a buffer, then echo it and terminate the script.
        ob_start();
            require($workpath . '/assets/templates/fullpage.php');
        echo ob_get_clean();

        // nvar_dump(get_included_files());

        exit;
    }

    function printDefinitions($area) {
        global $defs;
        $ndefs = $defs[$area];
        // Sort definitions alphabetically.
        ksort($ndefs);

                // Create new words array.
        $words = array();

        // Iterate through associative array
        foreach($ndefs as $key => $def) {

            // This is a process.
            //    1. Get first letter of the word and force it to uppercase.
            //    2. If there is no array for this letter, create it.
            //    3. Add the word to the array of this letter. As we are using array_push and have
            //       previously sorted the words list alphabetically, the resulting definitions in
            //       each of the per-letter arrays in $words will be in alphabetical order. We write
            //       the value in HTML so that it can just be echoed out later (notice the purifying
            //       a few steps back)
            $start = strtoupper(substr($key, 0, 1));
            
            if(empty($words[$start])) {
                $words[$start] = array();
            }

            array_push(
                $words[$start],
                '<p class="definition" id="' . $key . '"><span class="body"><strong><a href="#' . $key . '">'. $key . '</a></strong>: ' . $def . '</span></p>');
        }

        // Iterate through associative array, this time $words.
        //    1. Each $key is a letter of the alphabet, in uppercase, as created by the above process.
        //       We start by echoing this letter with a label and clickable link.
        //    2. For each value listed under this nested array, echo it after parsing any contained
        //       BBCode.
        foreach($words as $key => $list) {
            echo '<br /><hr /><h2 id="', $key, '" class="defblock"><a href="#', $key, '">', $key, '</a></h2>';
            foreach ($list as $word) {
                echo $word;
            }
        }
    }

    function renderDefRef($area, $definition, $label = "", $return = false) {
        global $defs;
        $def = $defs[$area][$definition];
        $def = htmlspecialchars(strip_tags($def));

        $label = ($label == "" ? $definition : $label);

        $string = "<span class=\"definition\" title=\"" . $def . "\">{$label}</span>";

        if($return === true)
            return $string;

        echo $string;
    }

    // Returns a policy statement, sometimes with HTML and sometimes with a link to read the full policy.
    function policyStatement($type, $statements, $addlink = false, $html = true) {
        $type = strtolower($type);
        global $$type;

        if(!is_array($statements))
            $statements = [$statements];

        $message = '';
        foreach($statements as $key => $statement) {
            $message .= ($key > 0 ? ' ' : '') . $$type[$statement];
        }

        return
            ($html ? '<p class="footnote"><strong>' : '') .
                ucfirst($type) . ':' . 
            ($html ? '</strong> ' : ' ')  .
                $message .
                ($addlink ? ' Read the full ' . $type . ' ' . 
                    ($type == 'privacy' ? 'policy ' : '') .
                printUrl(array('/?' . $type,'here'), 1, null, true) . '.' : '') .
            ($html ? '</p>' : '');
    }

    function countWords($page) {
        // Retrieve all HTML tags on the page but not the contents of those tags.
        preg_match_all('/<[\!\-\w\s\-\;="\'\.\/\\:#\?\&\,_]+>/', $page, $result);

        // Get the important stuff (aka, the whole tags).
        $data = $result[0];

        // Remove any duplicates.
        array_unique($data);

        // For each tag listed,
        foreach($data as $word) {
            // Get the tag's name
            $tag = preg_replace('/<(\/)?(\w+).+/', '$2', $word);

            // If the tag name is anything in the if, remove the entire tag.
            // Otherwise, replace the tag with a space.
            if(
                $tag === "sup"
            ) {
                $page = str_replace($word, "", $page);
            } else {
                $page = str_replace($word, " ", $page);
            }
        }

        // Remove unnecessary line breaks, etc. (3)
        // Replace multiple spaces with a single space. (1)
        // Removes 13 words which appear on any page. (1)
        // Remove spaces before solitary periods. (1)
        $page = str_ireplace("\r", "", $page); // Remove newline chars (part one)
        $page = str_ireplace("\n", "", $page); // Remove newline chars (part two)
        $page = str_ireplace("\t", "", $page); // Remove tabs
        $page = preg_replace('/\s+/', ' ', $page); // Replace multiple spaces with only one space.
        //$page = preg_replace('/Link to this page \| Last modified: \w+ \d+, \d+, \d\d:\d\d \w+ \(\+\d+\)/', '', $page);
        $page = preg_replace('/\s(\.\s)/', '$1', $page);

        // Trim any useless spaces from the start and end of the page.
        $page = trim(strip_tags($page));

        // Actually count the words, yay!
        $count = preg_match_all('/[\.\&\w\-\'\:\|\/\\\\()\+]+/', $page, $test);
        
        return $count;
    }

    function getUrl($url, $local = 1) {
        global $workurl, $workdomain, $sitedir;

        $pre = $workurl;

        if($local == 0)
            $pre = $workdomain . $sitedir . '/?redirect:::url=';
        
        if($local == 2) 
            $pre = $workdomain;

        return $pre . $url;
    }

    // input = array(url, label) or simply string containing url
    // refer = is this linking to an external website?
    // classes = classes to apply to the thing.
    $printUrl = "printUrl";
    function printUrl($input, $local = 1, $classes = null, $return = false) {
        // Default URL and href label to null.
        $url = null;
        $name = null;
        $target = "";

        // If input is an array, set variables appropriately. If it isn't, set both to the parameter given.
        if(is_array($input)) {
            $url = $input[0];
            $name = $input[1];
        } else {
            $url = $name = $input;
        }

        $url = getUrl($url, $local);

        // if(!is_array($input))
            // $name = $url;

        // work stuff out.
        $toPrint =
        '<a href="' . $url . '" title="Destination: ' . ($local == 0 ? '[EXTERNAL] ' : '') . $url . '"' .
            ($classes != null ? 'class="' . $classes . '"' : '') .
            ($local == 0 ? ' target="_blank"' : '') .
        '>' . $name . ($local == 0 ? ' [ext]' : '') .'</a>';

        // Echo URL.
        if($return === true)
            return $toPrint;

        echo $toPrint;
    }

    function addNavitem($navitem) {
        global $navitems;
        $navitems[] = $navitem;

        $navitems['pages'][] = $navitem[0];
    }

    function renderNavitem($link, $title, $level) {
        global $datapath, $sections, $pagedata;
        // Don't display if override parameter is set to false.
        if($level === false)
                return;

        $classes = 'navitem sub' . $level;

        // Remove trailing slashes
        if(substr($link, -1) == "/" && strlen($link) > 1)
            $link = substr($link, 0, -1);

        if($pagedata['navpath'] !== false)
            if(in_array($link, $pagedata['navpath']))
                $classes .= ' active';

        // if($datatree == $linktree)
            // $classes .= ' active';

        echo '<div class="', $classes, '">',
            printUrl(array($link,$title));

            if($datapath == $link && count($sections) !== 0) {
                echo '<div class="sections indent8">';
                foreach($sections as $section) {
                    echo '<div class="item">',
                        printUrl(array("{$link}#{$section[2]}", strip_tags($section[1]))),
                    '</div>';
                }
                echo '</div>';
            }


        echo '</div>';

    }

    function StartNewSection($level, $title = null, $id = null, $link = 1) {
        global $sections, $section_open, $pages, $datapath;

        if($level == 1 && isset($pages[$datapath]['header']))
            $title = $pages[$datapath]['header'];
        else
            $sections[] = array($level, $title, $id, $link);

        if($title == null)
            return;

        closeOpenSection();

        $section_open = true;

        // If not first-level, manual linebreak. This helps retain formatting when copying text.
        echo
            ( ($level > 1) ? '<br style="font-size:16px;" />' : '' ) . 
            '<div' .
            ( ($id !== null) ? (' id="' . $id . '"') : '' ) . // If there is an ID, insert it.
            ' class="section s'.$level.'"' .
            '><div class="header"><h'. $level .'>' .
            ( ( $id !== null )
                ?
                    ( ($link === 2) ? '<a href="#'. $id . '" title="Right click, copy link, share">'. $title . '</a>' : $title)
                :
                    $title
                ) . 
            '</h'. $level .'>' . 
            // If ID present and link type is 1, print a "Link to this section" link.
            ( ( $id !== null && $link === 1) ? '<br /><a href="#' . $id . '" title="Right click, copy link, share" class="section-' . $level . '">Link to this section</a>' : '') . 
            '</div>'
            ;
    }

    // Close the previous section if it's open.
    function closeOpenSection() {
        global $section_open;
        if($section_open) {
            $section_open = false;
            echo "</div>";
        }
    }

    /* return hit data for this page. */
    function readHits() {
        global $workpath, $datapath;
        $db = new SQLite3($workpath . "/assets/data/hits.sqlite");

        // If the table does not exist.
        $db->exec('CREATE TABLE IF NOT EXISTS hits (page VARCHAR(255) PRIMARY KEY NOT NULL, hitcount INT(10) DEFAULT 0)');

        $hitdata = array(
            'global'  => $db->querySingle('SELECT SUM(hitcount) FROM hits'),
            $datapath => $db->querySingle('SELECT hitcount FROM hits WHERE page = "' . $datapath . '"'),
        );

        $db->close();

        return $hitdata;
    }

    function logHit() {
        global $workpath, $hitlogged, $navitems, $datapath;

        // If hit logged or useragent is BOT, don't log hit.
        if($hitlogged || $_SERVER['HTTP_USER_AGENT'] === "BOT")
            return;

        $db = new SQLite3($workpath . "/assets/data/hits.sqlite");

        // If the table does not exist.
        $db->exec('CREATE TABLE IF NOT EXISTS hits (page VARCHAR(255) PRIMARY KEY NOT NULL, hitcount INT(10) DEFAULT 0)');

        if(in_array($datapath, $navitems['pages']))
            $db->exec('INSERT OR IGNORE INTO hits (page) VALUES ("' . $datapath . '"); UPDATE hits SET hitcount = hitcount + 1 WHERE page = "' . $datapath . '";');

        $db->close();

        $hitlogged = true;

        // If this is a milestone hit, create a download archive specifically for it.
        $hitdata = readHits();
        if(
            $hitdata['global'] == 100000 || 
            $hitdata['global'] == 250000 || 
            $hitdata['global'] == 500000 || 
            $hitdata['global'] == 1000000
        )
            createSiteDownload("MILESTONE ");
    }    

    // Some code borrowed and adapted from here:
    //     https://www.geeksforgeeks.org/how-to-zip-a-directory-in-php/
    function createSiteDownload($name = "") {
        global $workpath, $formatter;
        $zipname = $name . 'BLSAFETY.NET ' . datefmt_format($formatter[2], time()) . ".zip";
        $zippath = $workpath . '/../' . $zipname;

        if(!file_exists($zippath))
            Zip($workpath, $zippath, $zipname);

        return $zippath;
    }

    // Function taken and adapted from
    //    https://stackoverflow.com/questions/1334613/how-to-recursively-zip-a-directory-in-php
    function Zip($source, $destination, $internal_path)
    {
        if (!extension_loaded('zip') || !file_exists($source))
            return false;

        $zip = new ZipArchive();
        if (!$zip->open($destination, ZIPARCHIVE::CREATE))
            return false;

        $source = str_replace('\\', '/', realpath($source));

        $internal_path = str_replace('.zip', '', $internal_path);

        if (is_dir($source) === true)
        {
            $files = new RecursiveIteratorIterator(new RecursiveDirectoryIterator($source), RecursiveIteratorIterator::SELF_FIRST);

            foreach ($files as $file)
            {
                $file = str_replace('\\', '/', $file);

                // Ignore "." and ".." folders
                if(
                    in_array(substr($file, strrpos($file, '/')+1), array('.', '..', '.well-known')) ||
                    strpos($file, ".well-known") !== false
                )
                    continue;

                $file = realpath($file);

                if (is_dir($file) === true)
                {
                    $zip->addEmptyDir(str_replace($source . '/', $internal_path . '/', $file . '/'));
                }
                else if (is_file($file) === true)
                {
                    $zip->addFromString(str_replace($source . '/', $internal_path . '/', $file), file_get_contents($file));
                }
            }
        }
        else if (is_file($source) === true)
        {
            $zip->addFromString(basename($source), file_get_contents($source));
        }

        $zip->deleteName(".well-known");

        return $zip->close();
    }
?>